/*
 * Copyright (c) 2001 bobby
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the authors nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS `AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "services.h"

void getLogInfo(FILE *log, UserList *to, char *searchkey) {

  char  buffer[512];
  char  buff1[512];
  char  buff2[512];

  strcpy(buff2, searchkey); strtolower(buff2);

  uSend(LogServ, to, GetReply(RPL_LOGHEADER, to));
  while (sfgets(buffer, 1024, log) != NULL) {
    
    strcpy(buff1, buffer); strtolower(buff1);

    if (strstr(buff1, buff2)) {
      uSend(LogServ, to, GetReply(RPL_LOGINFO, to), buffer);
    }
  }
  uSend(LogServ, to, GetReply(RPL_LOGFOOTER, to));
}

void sendToLogServ (UserList *tmp, char **args, int numargs) {

  char *from = tmp->nick;

  if (!strcmp(args[0], "\001PING")) {
    if (numargs < 3)
      uSend(LogServ, tmp, GetReply(RPL_PING1, tmp), args[1]);
    else
      uSend(LogServ, tmp, GetReply(RPL_PING2, tmp), args[1], args[2]);
    return;
  }

  if (!isOper(tmp)) {
    uSend(LogServ, tmp, GetReply(ERR_NEEDOPER, tmp));
    return;
  }  

  if (!strcasecmp(args[0], "help") && numargs == 1)
    sendHelp(from, LogServ, NULL, 0);
  else if (!strcasecmp(args[0], "help"))
    sendHelp(from, LogServ, args, numargs);
  else
    msgtabCheck(LogServ, tmp, args, numargs);
}

void LshowlogCmd (UserList *nick, char **args, int numargs) {

  FILE *log;
  char key[512];
  int i;

  if (!strcasecmp(args[0], "nick")) {
    if (numargs < 2) {
      uSend(LogServ, nick, GetReply(ERR_BADPARAMS, nick));
      uSend(LogServ, nick, GetReply(RPL_MOREHELPCMD, nick), LogServ, "NICK");
      return;
    }
    log = fopen(LOGDIR "/nickserv.log", "r");
    if (!log) {
      uSend(LogServ, nick, "Lo siento, no se pudo accesar al log %s.", LOGDIR "/nickserv.log");
      return;
    }
  } else if (!strcasecmp(args[0], "chan")) {
    if (numargs < 2) {
      uSend(LogServ, nick, GetReply(ERR_BADPARAMS, nick));
      uSend(LogServ, nick, GetReply(RPL_MOREHELPCMD, nick), LogServ, "CHAN");
      return;
    }
    log = fopen(LOGDIR "/chanserv.log", "r");
    if (!log) {
      uSend(LogServ, nick, "Lo siento, no se pudo accesar al log %s.", LOGDIR "/chanserv.log");
      return;
    }
  } else if (!strcasecmp(args[0], "oper")) {
    if (numargs < 2) {
      uSend(LogServ, nick, GetReply(ERR_BADPARAMS, nick));
      uSend(LogServ, nick, GetReply(RPL_MOREHELPCMD, nick), LogServ, "OPER");
      return;
    }
    log = fopen(LOGDIR "/operserv.log", "r");
    if (!log) {
      uSend(LogServ, nick, "Lo siento, no se pudo accesar al log %s.", LOGDIR "/operserv.log");
      return;
    }
  } else if (!strcasecmp(args[0], "root")) {
    if (numargs < 2) {
      uSend(LogServ, nick, GetReply(ERR_BADPARAMS, nick));
      uSend(LogServ, nick, GetReply(RPL_MOREHELPCMD, nick), LogServ, "ROOT");
      return;
    }
    log = fopen(LOGDIR "/root.log", "r");
    if (log == NULL) {
      uSend(LogServ, nick, "Lo siento, no se pudo accesar al log %s.", LOGDIR "/root.log");
      return;
    }
  } else {
    uSend(LogServ, nick, GetReply(ERR_UNKNOWNCOMMAND, nick), args[0]);
    uSend(LogServ, nick, GetReply(RPL_MOREHELP, nick), LogServ);
    return;
  }

  memset(key, '\0', 511);
  for (i = 1; i < numargs; i++) {
    strcat(key, args[i]);
    if (!(i + 1 == numargs))
      strcat(key, " ");
  }
  getLogInfo(log, nick, key);
  fclose(log);
}

void LcommentCmd (UserList *nick, char **args, int numargs) {

  char  comment[1024];
  int   i;

  if (numargs < 3) {
    uSend(LogServ, nick, GetReply(ERR_BADPARAMS, nick));
    uSend(LogServ, nick, GetReply(RPL_MOREHELPCMD, nick), LogServ, "COMMENT");
    return;
  }

  memset(comment, '\0', 1024);
  for (i = 2; i < numargs; i++) {
    strcat(comment, args[i]);
    strcat(comment, " ");
  }

  if (index(args[1], '#')) {
    sSend(":%s GLOBOPS :COMMENT by %s for channel %s : %s", LogServ, nick->nick, args[1], comment);
    logDump(chanlog, "COMMENT by %s for channel %s : %s", nick->nick, args[1], comment);
  } else {
    sSend(":%s GLOBOPS :COMMENT by %s for nickname %s : %s", LogServ, nick->nick, args[1], comment);
    logDump(nicklog, "COMMENT by %s for nickname %s : %s", nick->nick, args[1], comment);
  }
}
