/*
 * Copyright (c) 2001 bobby
 * Copyright (c) 1999-2000 GZ
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the authors nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS `AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "services.h"

struct Message NSmsgtab[] = {
  { "info",        NinfoCmd,        0, 0, 0, 1 },
  { "register",    NregisterCmd,    0, 0, 0, 3 },
  { "ghost",       NghostCmd,       0, 0, 0, 3 },
  { "listops",     NlistopsCmd,     0, 0, 0, 1 },
  { "release",     NreleaseCmd,     0, 0, 0, 3 },
  { "id",          NidentifyCmd,    NEEDACTNICK, 0, 0, 2 },
  { "identify",    NidentifyCmd,    NEEDACTNICK, 0, 0, 2 },
  { "sidentify",   NidentifyCmd,    NEEDACTNICK, 0, 0, 2 },
  { "activate",    NactivateCmd,    NEEDREGNICK, 0, 0, 2 },
  { "set",         NsetCmd,         NEEDID, 0, 0, 2 },
  { "addmask",     NaddMaskCmd,     NEEDID, 0, 0, 1 },
  { "drop",        NdropCmd,        NEEDID, 0, 0, 2 },
  { "access",      NaccessCmd,      NEEDID, 0, 0, 2 },
  { "delete",      NdeleteCmd,      NEEDACTNICK|NEEDOPER,0, 0, 2 },
  { "list",        NlistCmd,        NEEDOPER, 0, 0, 2 },
  { "getpass",     NgetpassCmd,     NEEDOPER, 0, 0, 2 },
  { "getkey",      NgetkeyCmd,      NEEDOPER, 0, 0, 2 },
  { "setflags",    NsetflagsCmd,    NEEDSERVOP, 0, 0, 3 },
  { "banish",      NbanishCmd,      NEEDROOT, 0, 0, 2 },
  { "getrealpass", NgetrealpassCmd, NEEDROOT, 0, 0, 2 },
  { "save",        NsaveCmd,        NEEDROOT, 0, 0, 1 },
#ifdef ANALYSE_FUNCTIONS
  { "debug",       NSdebug,         NEEDROOT, 0, 0, 0 },
#endif
  { NULL, NULL, 0,0,0,0 }
};

struct Message CSmsgtab[] = {
  { "acc",        CaccessCmd,     0, 0, 0, 2 },
  { "access",     CaccessCmd,     0, 0, 0, 2 },
  { "info",       CinfoCmd,       0, 0, 0, 2 },
  { "register",   CregisterCmd,   NEEDID, 0, 0, 2 },
  { "addop",      CaddopCmd,      NEEDID, 0, 0, 3 },
  { "addak",      CaddakCmd,      NEEDID, 0, 0, 3 },
  { "delop",      CdelopCmd,      NEEDID, 0, 0, 3 },
  { "delak",      CdelakCmd,      NEEDID, 0, 0, 3 },
  { "listop",     ClistopCmd,     NEEDACTNICK, 0, 0, 2 },
  { "listak",     ClistakCmd,     NEEDACTNICK, 0, 0, 2 },
  { "drop",       CdropCmd,       NEEDID, 0, 0, 2 },
  { "op",         CopCmd,         NEEDID, 0, 0, 2 },
  { "deop",       CdeopCmd,       NEEDID, 0, 0, 2 },
  { "mlock",      CmodelockCmd,   NEEDID, 0, 0, 3 },
  { "modelock",   CmodelockCmd,   NEEDID, 0, 0, 3 },
  { "restrict",   CrestrictCmd,   NEEDID, 0, 0, 3 },
  { "topiclock",  CtopiclockCmd,  NEEDID, 0, 0, 3 },
  { "set",        CsetCmd,        NEEDID, 0, 0, 3 },
  { "massdeop",   CmassdeopCmd,   NEEDACTNICK, 0, 0, 2 },
  { "masskick",   CmasskickCmd,   NEEDACTNICK, 0, 0, 2 },
  { "modeclear",  CmodeclearCmd,  NEEDACTNICK, 0, 0, 2 },
  { "unban",      CunbanCmd,      NEEDID, 0, 0, 3 },
  { "invite",     CinviteCmd,     NEEDID, 0, 0, 2 },
  { "accesslist", CaccesslistCmd, NEEDID, 0, 0, 1 },
  { "listopl",    ClistoplCmd,    NEEDACTNICK, 0, 0, 2 },
  { "list",       ClistCmd,       NEEDOPER, 0, 0, 3 },
  { "delete",     CdeleteCmd,     NEEDSERVOP, 0, 0, 2 },
  { "clist",      CclistCmd,      NEEDSERVOP, 0, 0, 2 },
  { "close",      CcloseCmd,      NEEDSERVOP, 0, 0, 3 },
  { "save",       CsaveCmd,       NEEDROOT, 0, 0, 1 },
#ifdef ANALYSE_FUNCTIONS
  { "debug",      CSdebug,        NEEDROOT, 0, 0, 0 },
#endif
  { NULL, NULL, 0,0,0,0 }
};

struct Message OSmsgtab[] = {
  { "jinx",      OjinxCmd,      NEEDOPER, 0, 0, 2 },
  { "list",      OlistCmd,      NEEDOPER, 0, 0, 2 },
  { "map",       OmapCmd,       NEEDOPER, 0, 0, 1 },
  { "trigger",   OtriggerCmd,   NEEDOPER, 0, 0, 2 },
  { "cloneset",  OclonesetCmd,  NEEDOPER, 0, 0, 2 },
  { "mode",      OmodeCmd,      NEEDOPER, 0, 0, 3 },
  { "uptime",    OuptimeCmd,    NEEDOPER, 0, 0 ,1 },
  { "logonnews", OlogonnewsCmd, NEEDOPER, 0, 0, 2 },
  { "opernews",  OopernewsCmd,  NEEDOPER, 0, 0, 2 },
  { "listops",   OlistopsCmd,   NEEDOPER, 0, 0, 1 },
  { "masskill",  OmkillCmd,     NEEDSERVOP, 0, 0, 2 },
  { "mkill",     OmkillCmd,     NEEDSERVOP, 0, 0, 2 },
  { "permjinx",  OpermjinxCmd,  NEEDSERVOP, 0, 0, 2 },
  { "hostjinx",  OhostjinxCmd,  NEEDSERVOP, 0, 0, 2 },
  { "count",     OcountCmd,     NEEDSERVOP, 0, 0, 2 },
  { "broadcast", ObroadcastCmd, NEEDSERVOP, 0, 0, 2 },
  { "ekill",     OekillCmd,     NEEDSERVOP, 0, 0, 2 },
  { "jupe",      OjupeCmd,      NEEDSERVOP, 0, 0, 2 },
  { "jupiter",   OjupeCmd,      NEEDSERVOP, 0, 0, 2 },
  { "sopops",    OsopopsCmd,    NEEDSERVOP, 0, 0, 2 },
  { "raw",       OrawCmd,       NEEDROOT, 0, 0, 2 },
  { "inject",    OinjectCmd,    NEEDROOT, 0, 0, 2 },
  { "enforce",   OenforceCmd,   NEEDROOT, 0, 0, 1 },
  { "shutdown",  OshutdownCmd,  NEEDROOT, 0, 0, 1 },
  { "svsnoop",   OsvsnoopCmd,   NEEDROOT, 0, 0, 3 },
  { "allocstat", OallocstatCmd, NEEDROOT, 0, 0, 1 },
  { "restart",   OrestartCmd,   NEEDROOT, 0, 0, 1 },
  { "sync",      OsyncCmd,      NEEDROOT, 0, 0, 1 },
  { "rehash",    OrehashCmd,    NEEDROOT, 0, 0, 1 },
  { "rootops",   OrootopsCmd,   NEEDROOT, 0, 0, 2 },
#ifdef ANALYSE_FUNCTIONS
  { "debug",     OSdebug,       NEEDROOT, 0, 0, 0 },
#endif
  { NULL, NULL, 0,0,0,0 }
};

struct Message MSmsgtab[] = {
  { "del",      MdeleteCmd,   NEEDACTNICK|NEEDID, 0, 0, 2 },
  { "delete",   MdeleteCmd,   NEEDACTNICK|NEEDID, 0, 0, 2 },
  { "set",      MsetCmd,      NEEDACTNICK|NEEDID, 0, 0, 2 },
  { "retrieve", MretrieveCmd, NEEDACTNICK|NEEDID, 0, 0, 1 },
  { "forward",  MforwardCmd,  NEEDACTNICK|NEEDID, 0, 0, 3 },
  { "list",     MlistCmd,     NEEDACTNICK|NEEDID, 0, 0, 1 },
  { "send",     MsendCmd,     NEEDACTNICK|NEEDID, 0, 0, 3 },
  { "csend",    McsendCmd,    NEEDACTNICK|NEEDID, 0, 0, 4 },
  { "read",     MreadCmd,     NEEDACTNICK|NEEDID, 0, 0, 2 },
#ifdef ANALYSE_FUNCTIONS
  { "debug",    MSdebug,      NEEDROOT, 0, 0, 0 },
#endif
  { NULL, NULL, 0,0,0,0 }
};

struct Message LSmsgtab[] = {
  { "chan",    LshowlogCmd, NEEDOPER, 0, 0, 2 },
  { "nick",    LshowlogCmd, NEEDOPER, 0, 0, 2 },
  { "oper",    LshowlogCmd, NEEDOPER, 0, 0, 2 },
  { "comment", LcommentCmd, NEEDSERVOP, 0, 0, 3 },
  { "root",    LshowlogCmd, NEEDROOT, 0, 0, 2 },
#ifdef ANALYSE_FUNCTIONS
  { "debug",   LSdebug,     NEEDROOT, 0, 0, 0 },
#endif
  { NULL, NULL, 0,0,0,0 }
};

void msgtabCheck(char *Service, UserList *tmp, char **args, int numargs) {
  
  int i,registered, activated;
  int oper, servop, root, id;
  long timenow;
#ifdef ANALYSE_FUNCTIONS
  long before, ticks;
#endif
  struct Message *tabpointer;

  i = registered = activated = 0;
  oper = servop = root = id = 0;
  timenow = time(NULL);

  if (DoFloodCheck(tmp)) {
    tmp->lastmsg = timenow;
    return;
  } else
    tmp->lastmsg = timenow;

  if (tmp->reg) {
    registered = 1;

    if (tmp->reg->flags & NACTIVE) {
      activated = 1;
      if (isServop(tmp))
        servop = 1;
      if (isRoot(tmp))
        root = 1;
    }
  }
 
  if (isOper(tmp))
    oper = 1;

  if (tmp->caccess > 2)
    id = 1;

  if (!strcmp(Service, NickServ))
    tabpointer = NSmsgtab;
  else if (!strcmp(Service, ChanServ))
    tabpointer = CSmsgtab;
  else if (!strcmp(Service, OperServ))
    tabpointer = OSmsgtab;
  else if (!strcmp(Service, MemoServ))
    tabpointer = MSmsgtab;
  else if (!strcmp(Service, LogServ))
    tabpointer = LSmsgtab;
  else
    return;

  while (tabpointer[i].cmd) {
    if (!strcasecmp(tabpointer[i].cmd, args[0])) {
      if ((tabpointer[i].flag & NEEDREGNICK) && !registered) {
        uSend(Service, tmp, GetReply(ERR_NEEDREGNICK, tmp));
        return;
      } else if ((tabpointer[i].flag & NEEDACTNICK) && !activated) {
        uSend(Service, tmp, GetReply(ERR_NEEDACTNICK, tmp));
        return;
      } else if((tabpointer[i].flag & NEEDROOT) && !root) {
        uSend(Service, tmp, GetReply(ERR_NEEDSRA, tmp));
        return;
      } else if ((tabpointer[i].flag & NEEDSERVOP) && !servop) {
        uSend(Service, tmp, GetReply(ERR_NEEDSERVOP, tmp));
        return;
      } else if ((tabpointer[i].flag & NEEDOPER) && !oper) {
        uSend(Service, tmp, GetReply(ERR_NEEDOPER, tmp));
        return;
      } else if ((tabpointer[i].flag & NEEDID) && !id) {
        uSend(Service, tmp, GetReply(ERR_NEEDID, tmp));
        return;
      }

      if (tabpointer[i].params > numargs && tabpointer[i].params != 0) {
        uSend(Service, tmp, GetReply(ERR_BADPARAMS, tmp));
        uSend(Service, tmp, GetReply(RPL_MOREHELPCMD, tmp), Service, tabpointer[i].cmd);
        return;
      }
        
#ifdef ANALYSE_FUNCTIONS
      before = GetClock();
#endif
      tabpointer[i].func(tmp, args, numargs);
#ifdef ANALYSE_FUNCTIONS
      tabpointer[i].count++;
      ticks = (GetClock() - before);
      tabpointer[i].totaltime += ticks;
#endif
      return;
    }
    i++;
  }
  
  uSend(Service, tmp, GetReply(ERR_UNKNOWNCOMMAND, tmp), args[0]);
  uSend(Service, tmp, GetReply(RPL_MOREHELP, tmp), Service);
  return;
}

#ifdef ANALYSE_FUNCTIONS
void NSdebug(UserList *tmp, char **args, int numargs) {
                        
  long average;
  int i = 0;

  uSend(NickServ, tmp, GetReply(RPL_DEBUGHEADER, tmp), NickServ);
  uSend(NickServ, tmp, GetReply(RPL_DEBUGLINE1, tmp), "Command", "Count", "Total", "Average");
                                
  while (NSmsgtab[i].cmd) {
    if (NSmsgtab[i].count > 0)
      average = NSmsgtab[i].totaltime / NSmsgtab[i].count;
    else
      average = 0;

    uSend(NickServ, tmp, GetReply(RPL_DEBUGLINE2, tmp), NSmsgtab[i].cmd, NSmsgtab[i].count, NSmsgtab[i].totaltime, average);
    i++;
  }

  uSend(NickServ, tmp, GetReply(RPL_DEBUGFOOTER, tmp));
  return;                        
}

void OSdebug(UserList *tmp, char **args, int numargs) {
                        
  long average;
  int i = 0;

  uSend(OperServ, tmp, GetReply(RPL_DEBUGHEADER, tmp), OperServ);
  uSend(OperServ, tmp, GetReply(RPL_DEBUGLINE1, tmp), "Command", "Count", "Total", "Average");

  while (OSmsgtab[i].cmd) {
    if (OSmsgtab[i].count > 0)
      average = OSmsgtab[i].totaltime / OSmsgtab[i].count;
    else
      average = 0;
    uSend(OperServ, tmp, GetReply(RPL_DEBUGLINE2, tmp), OSmsgtab[i].cmd, OSmsgtab[i].count, OSmsgtab[i].totaltime, average);
    i++;
  }
  uSend(OperServ, tmp, GetReply(RPL_DEBUGFOOTER, tmp));
  return;                        
}

void MSdebug(UserList *tmp, char **args, int numargs) {
                        
  long average;
  int i = 0;

  uSend(MemoServ, tmp, GetReply(RPL_DEBUGHEADER, tmp), MemoServ);
  uSend(MemoServ, tmp, GetReply(RPL_DEBUGLINE1, tmp), "Command", "Count", "Total", "Average");
                                
  while (MSmsgtab[i].cmd) {
    if (MSmsgtab[i].count > 0)
      average = MSmsgtab[i].totaltime / MSmsgtab[i].count;
    else
      average = 0;
    uSend(MemoServ, tmp, GetReply(RPL_DEBUGLINE2, tmp), MSmsgtab[i].cmd, MSmsgtab[i].count, MSmsgtab[i].totaltime, average);
    i++;
  }
                        
  uSend(MemoServ, tmp, GetReply(RPL_DEBUGFOOTER, tmp));
  return;                        
}

void CSdebug(UserList *tmp, char **args, int numargs) {
                       
  long average;
  int i = 0;

  uSend(ChanServ, tmp, GetReply(RPL_DEBUGHEADER, tmp), ChanServ);
  uSend(ChanServ, tmp, GetReply(RPL_DEBUGLINE1, tmp), "Command", "Count", "Total", "Average");
                                
  while (CSmsgtab[i].cmd) {
    if (CSmsgtab[i].count > 0)
      average = CSmsgtab[i].totaltime / CSmsgtab[i].count;
    else
      average = 0;

    uSend(ChanServ, tmp, GetReply(RPL_DEBUGLINE2, tmp), CSmsgtab[i].cmd, CSmsgtab[i].count, CSmsgtab[i].totaltime, average);
    i++;
  }
                        
  uSend(ChanServ, tmp, GetReply(RPL_DEBUGFOOTER, tmp));
  return;                        
}

void LSdebug(UserList *tmp, char **args, int numargs) {
                       
  long average;
  int i = 0;

  uSend(LogServ, tmp, GetReply(RPL_DEBUGHEADER, tmp), LogServ);
  uSend(LogServ, tmp, GetReply(RPL_DEBUGLINE1, tmp), "Command", "Count", "Total", "Average");
                                
  while (LSmsgtab[i].cmd) {
    if (LSmsgtab[i].count > 0)
      average = LSmsgtab[i].totaltime / LSmsgtab[i].count;
    else
      average = 0;

    uSend(LogServ, tmp, GetReply(RPL_DEBUGLINE2, tmp), LSmsgtab[i].cmd, LSmsgtab[i].count, LSmsgtab[i].totaltime, average);
    i++;
  }
                        
  uSend(LogServ, tmp, GetReply(RPL_DEBUGFOOTER, tmp));
  return;                        
}

long GetClock(void) {
         
  long clocks;
  struct timeval *time;
  struct timezone *tz;
                                
  time = &fcountt;
  tz = &fcountz;
  gettimeofday(time, tz); 

  clocks = fcountt.tv_sec*1000000 + fcountt.tv_usec;
  return clocks;
}
#endif  

int DoFloodCheck(UserList *tmp) {
       
  long timenow = time(NULL);

  if (tmp->flooder == 0) {
    if (tmp->floodbytes < 7500) {
      if ((timenow - tmp->lastmsg) > 10)
        tmp->floodbytes = 0;
      return 0;
    } else {
      tmp->flooder = 1;
      return 1;
    }
  }
        
  if ((timenow - tmp->lastmsg) < 10)
    return 1;
  else {  
    tmp->floodbytes /= 2;
    if (tmp->floodbytes < 7500)
      tmp->flooder=0;
  }
  return 0;   
}
