#ifndef __NICKSERV_H
#define __NICKSERV_H

/* online flags: */
#define NISOPER    0x0001
#define NISAWAY    0x0002
#define NISHELPOP  0x0004
#define NISONLINE  0x0008
#define NISADMIN   0x0010
#define NISPLUSX   0x0020
#define NISINVIS   0x0040
#define NISIDENT   0x0080
#define NISCOADMIN 0x0100

/* NICKNAME FLAGS -- FIRST SLOT FILLED */
#define NKILL     0x00000001
#define NVACATION 0x00000002
#define NHOLD     0x00000004
#define NACTIVE   0x00000008
#define NREG      0x00000010
#define NSERVOP   0x00000020
#define NROOT     0x00000040
#define NEMAIL    0x00000080
#define NBANISH   0x00000100
#define NDEACC    0x00000200
#define NOPER     0x00000400
#define NGOREAN   0x00000800
#define NBOT      0x00001000
#define NREGBOT   0x00002000
#define NNETADM   0x00004000
#define NMARKED   0x00008000
#define NNOOP     0x00010000
#define NHIDE     0x00020000

/* NICKNAME FLAGS -- SECOND SLOT */
#define NROLEPLAY  0x00000001
#define NSTICKYOP  0x00000002
#define NNOPWWARN  0x00000004
#define NMSGMODE   0x00000008
#define NSIDENTIFY 0x00000020

/*
 * access list for a nick 
 */
typedef struct naccesslist_struct nAccessList;

struct naccesslist_struct {
  int index;
  char            mask[71]; /* their mask */
  nAccessList    *next;   /* The next mask. */
  nAccessList    *previous;
};

/*
 * Linked nicks for remote-ID (for now, more is likely to follow) -GZ
 */

typedef struct nlinkednicks_struct nLinkList;

struct nlinkednicks_struct {
  char    nick[NICKLEN+1];
  nLinkList     *next;  
  nLinkList     *previous;
};

/*
 * ===DOC===
 * This is 'better' docs of whats in the nicklist_struct structure:
 * nick: the persons nick
 * user: the persons username
 * host: the persons host.
 * timestamp: the time they signed on, also saved as their last seen time.
 * timereg: when their nick was registered, NULL if not registered.
 * amasks: how many access masks do we have, -1 if not registered.
 *     flags set on this nick, i.e. REGISTERED, HELD, KILL etc.
 * *AccessList: a pointer to a linked list of access list info for this
 *     person
 * *ChanList: a pointer to a linked list of the channels this person is on
 *     (this makes life somuch easier, trust me, :>)
 * next: the next (online) nickname.
 */

struct regnicklist_struct {
  char            nick[NICKLEN];
  char            user[USERLEN];
  char            host[HOSTLEN];
  char            realname[51];
  char            password[16];
  char            password2[12];
  char            url[128];
  char            email[EMAILLEN];
  int   language;
  time_t          timestamp;  /* Signon time (again) (Doubles as last seen) */
  time_t          timereg;  /* when this was registered. */
  time_t          lastpwchange;
  int             amasks; /* number of masks in access list */
  unsigned int    flags;  /* nick flags */
  unsigned int    flags2; /* additional flags */
  unsigned int    chans;
  char info[200]; /* users info, neat feature */
  nAccessList    *firstAcc, *lastAcc;
  MemoBox *memos;
  RegNickList   *next, *previous, *hashnext, *hashprev;
};

struct userlist_struct {
  char nick[NICKLEN];
  char user[USERLEN], host[HOSTLEN];
  char server[51];
  char realname[51];
  unsigned int oflags, floodlevel, caccess;
  unsigned int floodbytes;
  int linked; /* Number of linked nicks */
  int flooder;
  int floodwarn;
  int failid;
  int remotefailid;
  int resynch;
  time_t lastmsg;
  RegNickList *reg;
  nLinkList      *firstLink, *lastLink;
  UserList *next, *previous, *hashnext, *hashprev;
};

/* Global functions */

void        addNewUser(char **, int);
void        addNewUserSN(char **, int);
void        remUser(char *);
void        sendToNickServ(UserList *, char **, int);
void        setMode(char *, char *);
void        changeNick(char *, char *, char *);
UserList    *getNickData(char *);
RegNickList *getRegNickData(char *);
void        setFlags(char *, int, char);
void        cleanNickList(void);
int         isOper(UserList *);
int         isAdmin(UserList *);
int         isCoAdmin(UserList *);
int         isServop(UserList *);
int         isRoot(UserList *);
int         isNetAdmin(UserList *);
int         checkAccess(char *, char *, RegNickList *);
void        addAccessMask(char *, UserList *);
void        delAccessMask(char *, UserList *);
int         isGhost(char *);
void        delGhost(char *);
int         addFlood(UserList *, int);
void        nDesynch(char *, char *);
void        addNick(UserList *);
void        addRegNick(RegNickList *);
void        delNick(UserList *);
void        delRegNick(RegNickList *);
void        initNickData(RegNickList *);
int         addAccItem(RegNickList *, char *);
int         delAccItem(RegNickList *, char *, char *);
void        syncNickData(time_t);

/* NickServ commands */
void NidentifyCmd(UserList *, char **, int);
void NsetCmd(UserList *, char **, int);
void NregisterCmd(UserList *, char **, int);
void NactivateCmd(UserList *, char **, int);
void NinfoCmd(UserList *, char **, int);
void NdropCmd(UserList *, char **, int);
void NaddMaskCmd(UserList *, char **, int);
void NaccessCmd(UserList *, char **, int);
void NreleaseCmd(UserList *, char **, int);
void NsetflagsCmd(UserList *, char **, int);
void NbanishCmd(UserList *, char **, int);
void NgetpassCmd(UserList *, char **, int);
void NgetkeyCmd(UserList *, char **, int);
void NgetrealpassCmd(UserList *, char **, int);
void NdeleteCmd(UserList *, char **, int);
void NlistCmd(UserList *, char **, int);
void NhardDeleteCmd(UserList *, char **, int);
void NghostCmd(UserList *, char **args, int);
void NsaveCmd(UserList *, char **args, int);
void NlistopsCmd(UserList *, char **, int);

#endif /* __NICKSERV_H */
