/***************************************************************************
** This is where the more crucial defines are done. Don't start messing with
** these unless you are sure about what you are doing.
**                                      
** - bobby
***************************************************************************/

#ifndef __SERVICES_H
#define __SERVICES_H

#define sversion   1.10
#define RELEASE    "Development"
#define MAXSQ 500000
#define MAXRQ 500000

#undef  NOFORK
#define FORCE_CORE
#undef  DEBUG
#undef  DBDEBUG
#undef  CDEBUG
#undef  HASHDEBUG
#ifdef DEBUG
  void _debug(int level, char *msg, ...);
  #define Debug(x) _debug x
#else
  #define Debug(x) ;
#endif
#define DEBUG_ERROR  1
#define DEBUG_WARN   2
#define DEBUG_NOTICE 3
#define DEBUG_INFO   4
#define DEBUG_TCP    5
#undef  TELNET
#define ANALYSE_FUNCTIONS


#include <sys/types.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <ctype.h>
#include <time.h>
#include "options.h" /* Include the rest of our settings */
#include <sys/resource.h>

/* Define some standard lengths */
#define   NICKLEN 31
#define   USERLEN 15
#define   HOSTLEN 64
#define   EMAILLEN 80
#define   CHANLEN 33
#define   MEMOLEN 350

/* Hash table defines */
#define   NICKHASHSIZE 6007   /* How many nicks in the hash array */
#define   CHANHASHSIZE 2003   /* How many channels in the hash array */
#define   CHANUSERHASHSIZE 547    /* How many users in a channel */
#define   OPHASHSIZE 547      /* How many ops in a hash entry */
#define   VARHASHSIZE NICKHASHSIZE  /* Just leave this the same as nick */
#define   NICKCHANHASHSIZE 10
#define   CLONEHASHSIZE 2003

/* Define structures etc here */
#define A_AKILL 0x01
#define A_IGNORE 0x02
#define CLONE_KILLFLAG 0x01
#define CLONE_IGNOREFLAG 0x02
#define CLONE_PERMIGNORE 0x04
  
typedef struct userlist_struct UserList;

typedef struct service_struct {
  char            name[21];
  char            uname[11];
  char            host[67];
  char            rname[51];
  char            mode[10];
} Service;

typedef struct database_struct {
  FILE *ns;
  FILE *cs;
  FILE *os;
  FILE *nsreg;
  FILE *ms;
  FILE *w;
} database;

/* Global function prototypes */

void            readConf(void);
void            flushLogs(void);
void            expireChans(void);
void            expireNicks(void);
void            sshutdown(int);
void            srestart(void);
void     	disk_synch(void);
void            addUser(char *, char *, char *, char *, char *);
int     	listenOnPort(int);
int             ConnectToServer(char *, int);
struct hostent  *resolve(char *);
struct hostent  *lookupHost(char *);
struct hostent  *lookupIP(char *);
void            sSend(char *,...);
void    	uSend(char *, UserList *, char *, ...);
void    	sockSend(int, char *,...);
void    	killUser(Service, UserList *, char *);
void    	userMode(UserList *, char *);
void            logDump(FILE *, char *,...);
int             sendMail(char *, char *, char *, char *,...);
void            breakLine(char *);
void    	parseLine(char *);
void            segvdie(int);
int             cancel_timer(unsigned int);
unsigned int    timer(long, void (*func) (char *), char *, char *);
void    	checktimers(void);
int             breakString(int, char *, char **, char *);
int             match(char *, char *);
int             hasCodes(char *);
void            strtolower(char *);
void            doTfunc(char *);
char            *sfgets(char *, int, FILE *);
int     	check_match(char *);
int     	exp_match(char *, char *);
void            *smalloc(size_t);
void           	sfree(void *);
void      	loadPermJinxes();
void    	listAkills(char *from, char type, char *mask);
void    	addakill(long length, char *mask, char *by, char type, char *reason);
void    	removeakill(char *from, char *mask);
int     	findakill(char *mask);
void    	saveakills();
void    	loadakills();


/* Global variables */
int   		antiloop;
char    	timerbuff[512];
char    	debugbuff[512];
time_t    	debugtime;
time_t    	timertime;
char            myname[64];
char            mypass[33];
char            hostname[10][64];
char    	hostservername[64];
int             port[10];
int             server, synched;
Service   	services[6];
database        db;
FILE            *operlog, *nicklog, *chanlog, *corelog, *servout, *rootlog;
long            totalusers, mostusers, mostnicks, mostchans, mostmemos, startup;
unsigned int    AllocCalled, FreeCalled;
char            *OperServ, *NickServ, *ChanServ, *MemoServ, coreBuffer[512];
int             AccessLimit, OpLimit, AkickLimit, ChanLimit, NickLimit;
time_t          CTime;    /* current time, keep this in mind... */

typedef struct regnicklist_struct 	RegNickList;
typedef struct chanlist_struct 		ChanList;

struct regnicklist_struct 	*firstRegNick, *lastRegNick;
struct userlist_struct 		*firstUser, *lastUser;
struct nreglist_struct 		*firstReg, *lastReg;
struct chanlist_struct 		*firstChan, *lastChan;
struct regchanlist_struct 	*firstRegChan, *lastRegChan;
struct newsitem 		*firstNews, *lastNews;

#include "stuff.h"
#include "messages.h"
#include "news.h"
#include "memoserv.h"
#include "nickserv.h"
#include "chanserv.h"
#include "operserv.h"
#include "help.h"
#include "db.h"
#include "ban.h"
#include "replies.h"
#include "logserv.h"
#include "jinx.h"
#include "msgtab.h"
#include "hash.h"
#include "numeric.h"

UserHashEnt     UserHash[NICKHASHSIZE];
RegNickHashEnt  RegNickHash[NICKHASHSIZE];
ChanHashEnt     ChanHash[CHANHASHSIZE];
RegChanHashEnt  RegChanHash[CHANHASHSIZE];
CloneHashEnt    CloneHash[CLONEHASHSIZE];

#ifdef TELNET
#include "telnet.h"
  int telnetsock;
#endif

time_t nextNsync, nextCsync;
time_t nextMsync;

int cslock, mslock, nslock, pidlock;

char *LogServ;
char *GlobalMessenger;

/*
 * other MemoServ things you'll need:
 * MS_DEF_RCV_MAX = The *default* number of memos a max recieved field
 * will be set to
 * MS_RCV_MAX_HIGHEST = The highest a USER can set his max received field to 
 * MS_SENT_MAX = The most memos a person can send out per hour.
 *               IRCops are exempt, and forward groups and CSENDS count as 1.
 */

#define MS_DEF_RCV_MAX 50
#define MS_RCV_MAX_HIGHEST 200
#define MS_SENT_MAX 6
   
#define MAXFLOODLEVEL 100
#define NICKDROPTIME (14*24*3600)
#define CHANDROPTIME (14*24*3600)
#define PASSWARNTIME (60*24*3600)
#define MEMODROPTIME (10*24*3600)
#define SYNCTIME (15*60)                                /* Define time between synchs to disk */

/* SHOWSYNCH - globops each time databases are saves              */
/* SHOWSYNCHALL - notice to users showing how long until next save*/

#undef SHOWSYNCH
#ifdef SHOWSYNCH
    #undef SHOWSYNCHALL
#endif

/*
 * default clone flags:
 * 0 for none
 * CLONE_KILLFLAG to autokill for every host
 * CLONE_IGNOREFLAG to autoignore for every host
 * CLONE_PERMTRIGGER to make ALL triggers permanet (including unchaged)
 *
 * DEFHOSTCLONETRIGGER = Warning level for clones from *@host
 * DEFUSERCLONETRIGGER = Warning level for clones from user@host
 * DEFHOSTILEHOSTCLONES = Akill level for clones from *@host
 * DEFHOSTILEUSERCLONES = Akill level for clones from user@host
 */
   
#define DEFCLONEFLAGS 0
#define DEFHOSTCLONETRIGGER 5
#define DEFUSERCLONETRIGGER 5
#define DEFHOSTILEHOSTCLONES 35
#define DEFHOSTILEUSERCLONES 35

struct LangTable {
        char    language[20];
        int     numerics;
        struct  FormatTable *numtable;
};

struct FormatTable {
        int     numeric;
        char    *format;
};

char  retnum[2048];

int ToRecvQ(char *);
int FromRecvQ(void);
int ToSendQ(char *);
int FromSendQ(void);

struct  ArecvQ {        /* Incoming traffic. */
  int messages;
  int length;
  char  buffer[MAXRQ];    
} recvQ;

struct AsendQ {         /* Outgoing traffic. */
  int   messages;
  int   length;
  char  buffer[MAXSQ];    
} sendQ;

#define MAXREMOTEID 5 /* Define the maximum remote ID's per nick */

char  ftrack[512];    /* To store a function name here */

typedef struct ftrackstruct aftrackstruct;

struct  ftrackstruct {
  unsigned long starttime;
  char    function[512];
  aftrackstruct *next, *prev;
};

aftrackstruct ftracklist;

#endif /* __SERVICES_H */
