/*
 * Copyright (c) 1999-2000 GZ
 * Copyright (c) 1998 Bill Jones
 * Copyright (c) 1997 Max Byrd
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the authors nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS `AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "services.h"

/* new timer code 6/19/1997 --wd */

typedef struct timer_struct Timer;

struct timer_struct {
  time_t run;
  unsigned int id;
  void (*func)(char *);
  char *desc;
  char *args;
  Timer *next, *prev;
};

unsigned int timers = 0;
unsigned int timersused = 0;

time_t lastcheck;

void delTimer(Timer *);

Timer *firstTimer = NULL, *lastTimer = NULL;

unsigned int timer(long seconds, void (*func)(char *), char *desc, char *args) {
  Timer *newtimer;
  
  int goodvalue = 0;

  CTime = time(NULL);
  
  if(seconds <= 0) {
    func(args);
    return 0;
  }
  
  timersused++;
  timers++;

  if(!timers)  /* WHOOPS -- WE ROLLED OVER */
    timers++;

  while (!goodvalue) {
    goodvalue = 1;
    for (newtimer = firstTimer; newtimer; newtimer = newtimer->next) {
      if (newtimer->id == timers) {
        goodvalue = 0;
        break;
      }
    }
    if (!goodvalue)
    {
      timers++;
      if (!timers) /* STILL ROLLING OVER */
        timers++;
    }
  }

  newtimer = smalloc(sizeof(Timer));
  newtimer->id = timers;
  newtimer->func = func;
  newtimer->desc = desc;
  newtimer->args = smalloc(strlen(args)+1);
  strcpy(newtimer->args, args);
  newtimer->run = CTime + seconds;
  
  if(!firstTimer) {
    firstTimer = newtimer;
    newtimer->prev = NULL;
  }
  else {
    lastTimer->next = newtimer;
    newtimer->prev = lastTimer;
  }
  lastTimer = newtimer;
  newtimer->next = NULL;

  return timers;
}

void checktimers(void) {

  Timer *tmp, *next;
  CTime = time(NULL);

  if(lastcheck == CTime)
    return;
  
  for(tmp = firstTimer;tmp;tmp=next) {
    next=tmp->next;
    if(tmp->run <= CTime) {
      sprintf(timerbuff,"%s(%s)", tmp->desc, tmp->args);
      timertime = CTime;
      tmp->func(tmp->args);
      delTimer(tmp);
    }
  }
}

void delTimer(Timer *deltimer) {  
  timersused--;

  if(deltimer->prev)
    deltimer->prev->next = deltimer->next;
  else
    firstTimer = deltimer->next;

  if(deltimer->next)
    deltimer->next->prev = deltimer->prev;
  else
    lastTimer = deltimer->prev;

  sfree(deltimer->args);
  sfree(deltimer);
}

int cancel_timer(unsigned int id) {
  Timer *tmp;
  for(tmp=firstTimer;tmp;tmp=tmp->next) {
    if(tmp->id == id) {
      delTimer(tmp);
      return 1;
    }
  }
  return 0;
}
